require 'addressable'
require 'forwardable'
require 'sharepoint-http-auth'
require 'logger'
require 'sharepoint-ruby'

require_relative 'sharepoint_api/sharepoint_error'
require_relative 'sharepoint_api/version'
require_relative 'sharepoint_api/items'
require_relative 'sharepoint_api/file_info'
require_relative 'sharepoint_api/filename_cleaner'
require_relative 'sharepoint_api/file_system'
require_relative 'sharepoint_api/permissions'
require_relative 'sharepoint_api/users'
require_relative 'sharepoint_api/version_converter'

class SharepointApi
  extend Forwardable
  include SharepointApi::Items
  include SharepointApi::FileSystem
  include SharepointApi::Permissions
  include SharepointApi::Users

  class << self
    attr_writer :logger

    def logger
      @logger ||= begin
        logger = Logger.new($stdout)
        logger.level = Logger::WARN
        logger
      end
    end

    def log_as(method, exception_or_message, level: :info)
      message = if exception_or_message.respond_to?(:message)
                  exception_or_message.message
                else
                  exception_or_message
                end

      logger.send(level, "#{name}: #{method}: #{message}")
    end
  end

  attr_accessor :library_name
  attr_reader :site, :host

  def_delegators :site, :protocol
  def_delegators 'self.class', :logger, :log_as, :logger=

  def initialize(library_name: '', **config)
    @library_name = library_name
    @host = config.fetch(:host)
    @username = config.fetch(:username)
    @password = config.fetch(:password)
    @site_name = config[:site_name] || ''
    @ntlm = config[:ntlm] || false
    @verbose = config[:verbose] || false
    @login_name_extractor = config[:login_name_extractor]

    @site = build_connection
  end

  def site_relative_path(path, safe_quote: true)
    file_path = [
      library_name, path
    ].reject { |p| p.nil? || p == '' }.join('/')
    encode_path(file_path, safe_quote: safe_quote)
  end

  def server_relative_path(path)
    "/#{site_path}/#{site_relative_path(path)}"
  end

  def site_path
    @site.name
  end

  def base_path
    "#{protocol}://#{host}/#{site_path}"
  end

  def encode_path(path, safe_quote: true)
    path = Addressable::URI.encode(path)
    path.gsub!(/'/, '%27%27') if safe_quote
    path.gsub!('+', '%2B')
    path
  end

  def login_name(*args, **kwargs)
    @login_name_extractor.call(*args, **kwargs)
  end

  #######
  private
  #######

  def build_connection
    site = Sharepoint::Site.new(@host, @site_name)
    site.verbose = @verbose

    site.session = Sharepoint::HttpAuth::Session.new(site) if @ntlm
    site.session.authenticate(@username, @password) # custom STS?

    site
  end
end
